/*
 *	V8sim.c:	Straight forward implementation of a simulator for the V8
 *	Creation date:	01-10-2004
 *	Created by:	Allard van der Bas
 *	Contact:	pv2@digitalfluff.net
 *	Web address:	http://www.digitalfluff.net/pv2/
 *
 *	Changes: 24-11-2004
 *		Fixed flag handling for DEC / BTT / CMP / SUB.
 *
 *		Now passes the V8 stress test demo application.
 *
 *		 30-1-2005
 *		Added output functions for memory / processor status /
 *		disassembly.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "tables.h"

/*
 *  V8 main memory.
 */
unsigned char CPUmem[65535];

/*
 * Banks from FIRMWARE.BIN
 */
unsigned char bank0[3*4096];
unsigned char bank1[3*4096];
unsigned char bank2[3*4096];
unsigned char bank3[3*4096];
unsigned char bank4[3*4096];
unsigned char bank5[3*4096];
unsigned char bank6[3*4096];
unsigned char bank7[3*4096];
unsigned char bank8[3*4096];
unsigned char bank9[3*4096];

/*
 * V8 registers.
 */
unsigned char R0;
unsigned char R1;
unsigned char R2;
unsigned char R3;
unsigned char R4;
unsigned char R5;
unsigned char R6;
unsigned char R7;

unsigned char PSR;
unsigned short PC;
unsigned short mySP;

/*
 * Damn you windows and your f**ked up globals!
 * I always like to be send on a wild goose chase.
 * AvdB: 31-01-2005
 */
#define SP mySP

/*
 * defines for flag handling.
 */
#define IFLAG 0x08
#define NFLAG 0x04
#define CFLAG 0x02
#define ZFLAG 0x01

/*
 * Set the default stack to 007F
 */
#define STACK_DEFAULT 0x007F

void init_CPU() {
	R0 = 0;
	R1 = 0;
	R2 = 0;
	R3 = 0;
	R4 = 0;
	R5 = 0;
	R6 = 0;
	R7 = 0;

	PSR |= IFLAG;
	PC = 0;
	
	/* Set stack to 0x007F */
	SP = STACK_DEFAULT;
}

/*
 * WHAT is known about the mem map ?
 * 
 * This comes from www.maushammer.com/systems/dakotadigital/lcd-hardware.html
 *
 * F000 - FFFF registers.
 * FB00 - FFFF Unknown 16 registers ?
 * FA00 - FAFF memory map to somewhere ?
 * F920 - F9FF repeats of F900 - F91F
 * F900 - F903 SDRAM address for DMA
 * F800 - F8FF Unknown 16 registers ?
 * F800 - F801 SRAM address.
 * F802	       involved. 
 * F803        transfer direction 00 SRAM <- SDRAM 01 SRAM -> SDRAM
 * F804	       init transfer ?
 * F740 - F7FF repeats of F700 - F73F
 * F730 - F73F LCD screen parameters.
 * F72E - F72F Screen Power control 0x0002 is on.
 * F726 - F727 Bit mapped button register (SW6).
 * F722 - F723 Bit mapped button register (S1 SW1-SW5)
 * F718        USB related.
 * F716
 * F700 - F703 24 bit counter ?
 * C000 - EFFF Image of 0000 - 2FFF
 * 8000 - BFFF Image of 0000 - 3FFF
 * 4000 - 7FFF Image of 0000 - 3FFF
 * 1000 - 3FFF bank of code from FIRMWARE.BIN
 * 0090 - 0FFF Fixed bank of code from FIRMWARE.BIN
 * 0080 - 008F Interupt table 82/83 is USB interrupt vector.
 * 0000 - 007F Stack.
 */
unsigned char loadByte(unsigned short address) {
	return(CPUmem[address]);
}

void storeByte(unsigned short address,unsigned char value) {
	CPUmem[address] = value;
}

void setNZFlag(unsigned char value) {
	if (value == 0) {
		PSR |= ZFLAG;
	} else {
		PSR &= (0xFF - ZFLAG);
	}

	if (value & 0x80) {
		PSR |= NFLAG;
	} else {
		PSR &= (0xFF - NFLAG);
	}
}

void setCFlag(unsigned short value) {
	if (value > 0xFF) {
		PSR |= CFLAG;
	} else {
		PSR &= (0xFF - CFLAG);
	}
}

void setNOTCFlag(unsigned short value) {
	if (value > 0xFF) {
		PSR &= (0xFF - CFLAG);
	} else {
		PSR |= CFLAG;
	}
}

void INCR0() {
	unsigned short value;
	value = R0 + 1;
	R0++;
	setNZFlag(R0);
	setCFlag(value);
}

void INCR1() {
	unsigned short value;
	value = R1 + 1;
	R1++;
	setNZFlag(R1);
	setCFlag(value);
}

void INCR2() {
	unsigned short value;
	value = R2 + 1;
	R2++;
	setNZFlag(R2);
	setCFlag(value);
}

void INCR3() {
	unsigned short value;
	value = R3 + 1;
	R3++;
	setNZFlag(R3);
	setCFlag(value);
}

void INCR4() {
	unsigned short value;
	value = R4 + 1;
	R4++;
	setNZFlag(R4);
	setCFlag(value);
}

void INCR5() {
	unsigned short value;
	value = R5 + 1;
	R5++;
	setNZFlag(R5);
	setCFlag(value);
}

void INCR6() {
	unsigned short value;
	value = R6 + 1;
	R6++;
	setNZFlag(R6);
	setCFlag(value);
}

void INCR7() {
	unsigned short value;
	value = R7 + 1;
	R7++;
	setNZFlag(R7);
	setCFlag(value);
}

void ADCR0() {
	unsigned short tmp;

	tmp = R0 + R0;
	if (PSR & CFLAG) {
		tmp++;
	}
	R0 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R0);
}

void ADCR1() {
	unsigned short tmp;

	tmp = R0 + R1;
	if (PSR & CFLAG) {
		tmp++;
	}
	R0 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R0);
}

void ADCR2() {
	unsigned short tmp;

	tmp = R0 + R2;
	if (PSR & CFLAG) {
		tmp++;
	}
	R0 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R0);
}

void ADCR3() {
	unsigned short tmp;

	tmp = R0 + R3;
	if (PSR & CFLAG) {
		tmp++;
	}
	R0 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R0);
}

void ADCR4() {
	unsigned short tmp;

	tmp = R0 + R4;
	if (PSR & CFLAG) {
		tmp++;
	}
	R0 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R0);
}

void ADCR5() {
	unsigned short tmp;

	tmp = R0 + R5;
	if (PSR & CFLAG) {
		tmp++;
	}
	R0 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R0);
}

void ADCR6() {
	unsigned short tmp;

	tmp = R0 + R6;
	if (PSR & CFLAG) {
		tmp++;
	}
	R0 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R0);
}

void ADCR7() {
	unsigned short tmp;

	tmp = R0 + R7;
	if (PSR & CFLAG) {
		tmp++;
	}
	R0 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R0);
}

void TX0R0 () {
	R0 = R0;
	setNZFlag(R0);
}

void TX0R1 () {
	R0 = R1;
	setNZFlag(R0);
}

void TX0R2 () {
	R0 = R2;
	setNZFlag(R0);
}

void TX0R3 () {
	R0 = R3;
	setNZFlag(R0);
}

void TX0R4 () {
	R0 = R4;
	setNZFlag(R0);
}

void TX0R5 () {
	R0 = R5;
	setNZFlag(R0);
}

void TX0R6 () {
	R0 = R6;
	setNZFlag(R0);
}

void TX0R7 () {
	R0 = R7;
	setNZFlag(R0);
}

void ORR0 () {
	R0 |= R0;
	setNZFlag(R0);
}

void ORR1 () {
	R0 |= R1;
	setNZFlag(R0);
}

void ORR2 () {
	R0 |= R2;
	setNZFlag(R0);
}

void ORR3 () {
	R0 |= R3;
	setNZFlag(R0);
}

void ORR4 () {
	R0 |= R4;
	setNZFlag(R0);
}

void ORR5 () {
	R0 |= R5;
	setNZFlag(R0);
}

void ORR6 () {
	R0 |= R6;
	setNZFlag(R0);
}

void ORR7 () {
	R0 |= R7;
	setNZFlag(R0);
}

void ANDR0 () {
	R0 &= R0;
	setNZFlag(R0);
}

void ANDR1 () {
	R0 &= R1;
	setNZFlag(R0);
}

void ANDR2 () {
	R0 &= R2;
	setNZFlag(R0);
}

void ANDR3 () {
	R0 &= R3;
	setNZFlag(R0);
}

void ANDR4 () {
	R0 &= R4;
	setNZFlag(R0);
}

void ANDR5 () {
	R0 &= R5;
	setNZFlag(R0);
}

void ANDR6 () {
	R0 &= R6;
	setNZFlag(R0);
}

void ANDR7 () {
	R0 &= R7;
	setNZFlag(R0);
}

void XORR0() {
	R0 ^= R0;
	setNZFlag(R0);
}

void XORR1() {
	R0 ^= R1;
	setNZFlag(R0);
}

void XORR2() {
	R0 ^= R2;
	setNZFlag(R0);
}

void XORR3() {
	R0 ^= R3;
	setNZFlag(R0);
}

void XORR4() {
	R0 ^= R4;
	setNZFlag(R0);
}

void XORR5() {
	R0 ^= R5;
	setNZFlag(R0);
}

void XORR6() {
	R0 ^= R6;
	setNZFlag(R0);
}

void XORR7() {
	R0 ^= R7;
	setNZFlag(R0);
}

void ROLR0 () {
	unsigned short tmp;

	tmp = R0 << 1;
	if (PSR & CFLAG) {
		tmp |= 1;
	}
	R0 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R0);
}

void ROLR1 () {
	unsigned short tmp;

	tmp = R1 << 1;
	if (PSR & CFLAG) {
		tmp |= 1;
	}
	R1 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R1);
}

void ROLR2 () {
	unsigned short tmp;

	tmp = R2 << 1;
	if (PSR & CFLAG) {
		tmp |= 1;
	}
	R2 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R2);
}

void ROLR3 () {
	unsigned short tmp;

	tmp = R3 << 1;
	if (PSR & CFLAG) {
		tmp |= 1;
	}
	R3 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R3);
}

void ROLR4 () {
	unsigned short tmp;

	tmp = R4 << 1;
	if (PSR & CFLAG) {
		tmp |= 1;
	}
	R4 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R4);
}

void ROLR5 () {
	unsigned short tmp;

	tmp = R5 << 1;
	if (PSR & CFLAG) {
		tmp |= 1;
	}
	R5 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R5);
}

void ROLR6 () {
	unsigned short tmp;

	tmp = R6 << 1;
	if (PSR & CFLAG) {
		tmp |= 1;
	}
	R6 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R6);
}

void ROLR7 () {
	unsigned short tmp;

	tmp = R7 << 1;
	if (PSR & CFLAG) {
		tmp |= 1;
	}
	R7 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R7);
}

void RORR0 () {
	unsigned short tmp;
	int carry;
	
	carry = R0 & 0x01;
	tmp = R0 >> 1;
	if (PSR & CFLAG) {
		tmp |= 0x80;
	}
	R0 = tmp & 0xFF;
	setCFlag(carry << 8);
	setNZFlag(R0);
}

void RORR1 () {
	unsigned short tmp;
	int carry;
	
	carry = R1 & 0x01;
	tmp = R1 >> 1;
	if (PSR & CFLAG) {
		tmp |= 0x80;
	}
	R1 = tmp & 0xFF;
	setCFlag(carry << 8);
	setNZFlag(R1);
}

void RORR2 () {
	unsigned short tmp;
	int carry;
	
	carry = R2 & 0x01;
	tmp = R2 >> 1;
	if (PSR & CFLAG) {
		tmp |= 0x80;
	}
	R2 = tmp & 0xFF;
	setCFlag(carry << 8);
	setNZFlag(R2);
}

void RORR3 () {
	unsigned short tmp;
	int carry;
	
	carry = R3 & 0x01;
	tmp = R3 >> 1;
	if (PSR & CFLAG) {
		tmp |= 0x80;
	}
	R3 = tmp & 0xFF;
	setCFlag(carry << 8);
	setNZFlag(R3);
}

void RORR4 () {
	unsigned short tmp;
	int carry;
	
	carry = R4 & 0x01;
	tmp = R4 >> 1;
	if (PSR & CFLAG) {
		tmp |= 0x80;
	}
	R4 = tmp & 0xFF;
	setCFlag(carry << 8);
	setNZFlag(R4);
}

void RORR5 () {
	unsigned short tmp;
	int carry;
	
	carry = R5 & 0x01;
	tmp = R5 >> 1;
	if (PSR & CFLAG) {
		tmp |= 0x80;
	}
	R5 = tmp & 0xFF;
	setCFlag(carry << 8);
	setNZFlag(R5);
}

void RORR6 () {
	unsigned short tmp;
	int carry;
	
	carry = R6 & 0x01;
	tmp = R6 >> 1;
	if (PSR & CFLAG) {
		tmp |= 0x80;
	}
	R6 = tmp & 0xFF;
	setCFlag(carry << 8);
	setNZFlag(R6);
}

void RORR7 () {
	unsigned short tmp;
	int carry;
	
	carry = R7 & 0x01;
	tmp = R7 >> 1;
	if (PSR & CFLAG) {
		tmp |= 0x80;
	}
	R7 = tmp & 0xFF;
	setCFlag(carry << 8);
	setNZFlag(R7);
}

void DECR0 () {
	unsigned short tmp;
	tmp = R0 - 1;
	R0--;
	setNOTCFlag(tmp);
	setNZFlag(R0);
}

void DECR1 () {
	unsigned short tmp;
	tmp = R1 - 1;
	R1--;
	setNOTCFlag(tmp);
	setNZFlag(R1);
}

void DECR2 () {
	unsigned short tmp;
	tmp = R2 - 1;
	R2--;
	setNOTCFlag(tmp);
	setNZFlag(R2);
}

void DECR3 () {
	unsigned short tmp;
	tmp = R3 - 1;
	R3--;
	setNOTCFlag(tmp);
	setNZFlag(R3);
}

void DECR4 () {
	unsigned short tmp;
	tmp = R4 - 1;
	R4--;
	setNOTCFlag(tmp);
	setNZFlag(R4);
}

void DECR5 () {
	unsigned short tmp;
	tmp = R5 - 1;
	R5--;
	setNOTCFlag(tmp);
	setNZFlag(R5);
}

void DECR6 () {
	unsigned short tmp;
	tmp = R6 - 1;
	R6--;
	setNOTCFlag(tmp);
	setNZFlag(R6);
}

void DECR7 () {
	unsigned short tmp;
	tmp = R7 - 1;
	R7--;
	setNOTCFlag(tmp);
	setNZFlag(R7);
}

void SBCR0() {
	signed short tmp;
	tmp = R0 - R0;

	if (PSR & CFLAG) {
	} else {
		tmp = tmp - 1;
	}
	R0 = tmp & 0xFF;
	setNOTCFlag(tmp);
	setNZFlag(R0);
}

void SBCR1() {
	signed short tmp;

	tmp = R0 - R1;

	if (PSR & CFLAG) {
	} else {
		tmp = tmp - 1;
	}

	R0 = tmp & 0xFF;
	setNOTCFlag(tmp);
	setNZFlag(R0);
}

void SBCR2() {
	signed short tmp;
	tmp = R0 - R2;
	if (PSR & CFLAG) {
	} else {
		tmp = tmp - 1;
	}
	R0 = tmp & 0xFF;
	setNOTCFlag(tmp);
	setNZFlag(R0);
}

void SBCR3() {
	signed short tmp;
	tmp = R0 - R3;
	if (PSR & CFLAG) {
	} else {
		tmp = tmp - 1;
	}
	R0 = tmp & 0xFF;
	setNOTCFlag(tmp);
	setNZFlag(R0);
}

void SBCR4() {
	signed short tmp;
	tmp = R0 - R4;
	if (PSR & CFLAG) {
	} else {
		tmp = tmp - 1;
	}
	R0 = tmp & 0xFF;
	setNOTCFlag(tmp);
	setNZFlag(R0);
}

void SBCR5() {
	signed short tmp;
	tmp = R0 - R5;
	if (PSR & CFLAG) {
	} else {
		tmp = tmp - 1;
	}
	R0 = tmp & 0xFF;
	setNOTCFlag(tmp);
	setNZFlag(R0);
}

void SBCR6() {
	signed short tmp;
	tmp = R0 - R6;
	if (PSR & CFLAG) {
	} else {
		tmp = tmp - 1;
	}
	R0 = tmp & 0xFF;
	setNOTCFlag(tmp);
	setNZFlag(R0);
}

void SBCR7() {
	signed short tmp;
	tmp = R0 - R7;
	if (PSR & CFLAG) {
	} else {
		tmp = tmp - 1;
	}
	R0 = tmp & 0xFF;
	setNOTCFlag(tmp);
	setNZFlag(R0);
}

void ADDR0 () {
	unsigned short tmp;
	tmp = R0 + R0;
	R0 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R0);
}

void ADDR1 () {
	unsigned short tmp;
	tmp = R0 + R1;
	R0 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R0);
}

void ADDR2 () {
	unsigned short tmp;
	tmp = R0 + R2;
	R0 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R0);
}

void ADDR3 () {
	unsigned short tmp;
	tmp = R0 + R3;
	R0 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R0);
}

void ADDR4 () {
	unsigned short tmp;
	tmp = R0 + R4;
	R0 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R0);
}

void ADDR5 () {
	unsigned short tmp;
	tmp = R0 + R5;
	R0 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R0);
}

void ADDR6 () {
	unsigned short tmp;
	tmp = R0 + R6;
	R0 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R0);
}

void ADDR7 () {
	unsigned short tmp;
	tmp = R0 + R7;
	R0 = tmp & 0xFF;
	setCFlag(tmp);
	setNZFlag(R0);
}

void STP0() {
	PSR |= 0x01;
}

void STP1() {
	PSR |= 0x02;
}

void STP2() {
	PSR |= 0x04;
}

void STP3() {
	PSR |= 0x08;
}

void STP4() {
	PSR |= 0x10;
}

void STP5() {
	PSR |= 0x20;
}

void STP6() {
	PSR |= 0x40;
}

void STP7() {
	PSR |= 0x80;
}

// Z = ( (R0 & (1 << n)) == 0)
void BTT0() {
	unsigned char tmp;

	tmp = R0 & 0x01;
	setNZFlag(tmp);

	if (R0 & 0x01) {
		PSR &= (0xFF - ZFLAG);
	} else {
		PSR |= ZFLAG;
	}
}

void BTT1() {
	unsigned char tmp;

	tmp = R0 & 0x02;
	setNZFlag(tmp);

	if (R0 & 0x02) {
		PSR &= (0xFF - ZFLAG);
	} else {
		PSR |= ZFLAG;
	}
}

void BTT2() {
	unsigned char tmp;

	tmp = R0 & 0x04;
	setNZFlag(tmp);

	if (R0 & 0x04) {
		PSR &= (0xFF - ZFLAG);
	} else {
		PSR |= ZFLAG;
	}
}

void BTT3() {
	unsigned char tmp;

	tmp = R0 & 0x08;
	setNZFlag(tmp);

	if (R0 & 0x08) {
		PSR &= (0xFF - ZFLAG);
	} else {
		PSR |= ZFLAG;
	}
}

void BTT4() {
	unsigned char tmp;

	tmp = R0 & 0x10;
	setNZFlag(tmp);

	if (R0 & 0x10) {
		PSR &= (0xFF - ZFLAG);
	} else {
		PSR |= ZFLAG;
	}
}

void BTT5() {
	unsigned char tmp;

	tmp = R0 & 0x20;
	setNZFlag(tmp);

	if (R0 & 0x20) {
		PSR &= (0xFF - ZFLAG);
	} else {
		PSR |= ZFLAG;
	}
}

void BTT6() {
	unsigned char tmp;

	tmp = R0 & 0x40;
	setNZFlag(tmp);

	if (R0 & 0x40) {
		PSR &= (0xFF - ZFLAG);
	} else {
		PSR |= ZFLAG;
	}
}

void BTT7() {
	unsigned char tmp;

	tmp = R0 & 0x80;
	setNZFlag(tmp);

	if (R0 & 0x80) {
		PSR &= (0xFF - ZFLAG);
	} else {
		PSR |= ZFLAG;
	}
}

void CLP0() {
	PSR &= (0xFF - 0x01);
}

void CLP1() {
	PSR &= (0xFF - 0x02);
}

void CLP2() {
	PSR &= (0xFF - 0x04);
}

void CLP3() {
	PSR &= (0xFF - 0x08);
}

void CLP4() {
	PSR &= (0xFF - 0x10);
}

void CLP5() {
	PSR &= (0xFF - 0x20);
}

void CLP6() {
	PSR &= (0xFF - 0x40);
}

void CLP7() {
	PSR &= (0xFF - 0x80);
}

void T0XR0() {
	R0 = R0;
	setNZFlag(R0);
}

void T0XR1() {
	R1 = R0;
	setNZFlag(R1);
}

void T0XR2() {
	R2 = R0;
	setNZFlag(R2);
}

void T0XR3() {
	R3 = R0;
	setNZFlag(R3);
}

void T0XR4() {
	R4 = R0;
	setNZFlag(R4);
}

void T0XR5() {
	R5 = R0;
	setNZFlag(R5);
}

void T0XR6() {
	R6 = R0;
	setNZFlag(R6);
}

void T0XR7() {
	R7 = R0;
	setNZFlag(R7);
}

void CMPR0() {
	signed short test;
	test = R0 - R0;
	setNZFlag(test&0xFF);
	setNOTCFlag(test);
}

void CMPR1() {
	signed short test;
	test = R0 - R1;
	setNZFlag(test&0xFF);
	setNOTCFlag(test);
}

void CMPR2() {
	signed short test;
	test = R0 - R2;
	setNZFlag(test&0xFF);
	setNOTCFlag(test);
}

void CMPR3() {
	signed short test;
	test = R0 - R3;
	setNZFlag(test&0xFF);
	setNOTCFlag(test);
}

void CMPR4() {
	signed short test;
	test = R0 - R4;
	setNZFlag(test&0xFF);
	setNOTCFlag(test);
}

void CMPR5() {
	signed short test;
	test = R0 - R5;
	setNZFlag(test&0xFF);
	setNOTCFlag(test);
}

void CMPR6() {
	signed short test;
	test = R0 - R6;
	setNZFlag(test&0xFF);
	setNOTCFlag(test);
}

void CMPR7() {
	signed short test;
	test = R0 - R7;
	setNZFlag(test&0xFF);
	setNOTCFlag(test);
}

void PSHR0() {
	storeByte(SP,R0);
	SP--;
}

void PSHR1() {
	storeByte(SP,R1);
	SP--;
}

void PSHR2() {
	storeByte(SP,R2);
	SP--;
}

void PSHR3() {
	storeByte(SP,R3);
	SP--;
}

void PSHR4() {
	storeByte(SP,R4);
	SP--;
}

void PSHR5() {
	storeByte(SP,R5);
	SP--;
}

void PSHR6() {
	storeByte(SP,R6);
	SP--;
}

void PSHR7() {
	storeByte(SP,R7);
	SP--;
}

void POPR0() {
	SP++;
	R0 = CPUmem[SP];
}

void POPR1() {
	SP++;
	R1 = CPUmem[SP];
}

void POPR2() {
	SP++;
	R2 = CPUmem[SP];
}

void POPR3() {
	SP++;
	R3 = CPUmem[SP];
}

void POPR4() {
	SP++;
	R4 = CPUmem[SP];
}

void POPR5() {
	SP++;
	R5 = CPUmem[SP];
}

void POPR6() {
	SP++;
	R6 = CPUmem[SP];
}

void POPR7() {
	SP++;
	R7 = CPUmem[SP];
}

void BR00() {
	signed char offset;
	offset = CPUmem[PC++];
	
	if (PSR & 0x01) {
	} else {
		PC = PC + offset;
	}
}

void BR01() {
	signed char offset;
	offset = CPUmem[PC++];
	
	if (PSR & 0x02) {
	} else {
		PC = PC + offset;
	}
}

void BR02() {
	signed char offset;
	offset = CPUmem[PC++];
	
	if (PSR & 0x04) {
	} else {
		PC = PC + offset;
	}
}

void BR03() {
	signed char offset;
	offset = CPUmem[PC++];
	
	if (PSR & 0x08) {
	} else {
		PC = PC + offset;
	}
}

void BR04() {
	signed char offset;
	offset = CPUmem[PC++];
	
	if (PSR & 0x10) {
	} else {
		PC = PC + offset;
	}
}

void BR05() {
	signed char offset;
	offset = CPUmem[PC++];
	
	if (PSR & 0x20) {
	} else {
		PC = PC + offset;
	}
}

void BR06() {
	signed char offset;
	offset = CPUmem[PC++];
	
	if (PSR & 0x40) {
	} else {
		PC = PC + offset;
	}
}

void BR07() {
	signed char offset;
	offset = CPUmem[PC++];
	
	if (PSR & 0x80) {
	} else {
		PC = PC + offset;
	}
}

void BR10() {
	signed char offset;
	offset = CPUmem[PC++];
	
	if (PSR & 0x01) {
		PC = PC + offset;
	}
}

void BR11() {
	signed char offset;
	offset = CPUmem[PC++];
	
	if (PSR & 0x02) {
		PC = PC + offset;
	}
}

void BR12() {
	signed char offset;
	offset = CPUmem[PC++];
	
	if (PSR & 0x04) {
		PC = PC + offset;
	}
}

void BR13() {
	signed char offset;
	offset = CPUmem[PC++];
	
	if (PSR & 0x08) {
		PC = PC + offset;
	}
}

void BR14() {
	signed char offset;
	offset = CPUmem[PC++];
	
	if (PSR & 0x10) {
		PC = PC + offset;
	}
}

void BR15() {
	signed char offset;
	offset = CPUmem[PC++];
	
	if (PSR & 0x20) {
		PC = PC + offset;
	}
}

void BR16() {
	signed char offset;
	offset = CPUmem[PC++];
	
	if (PSR & 0x40) {
		PC = PC + offset;
	}
}

void BR17() {
	signed char offset;
	offset = CPUmem[PC++];
	
	if (PSR & 0x80) {
		PC = PC + offset;
	}
}

void USR() {
	/* No Clue */
}

void INT0 () {
	/* Not implemented */
}

void INT1 () {
	/* Not implemented */
}

void INT2 () {
	/* Not implemented */
}

void INT3 () {
	/* Not implemented */
}

void INT4 () {
	/* Not implemented */
}

void INT5 () {
	/* Not implemented */
}

void INT6 () {
	/* Not implemented */
}

void INT7 () {
	/* Not implemented */
}

void USR2() {
	/* No Clue */
}

void RSP() {
	SP = STACK_DEFAULT;
}

void RTS() {
	SP++;
	PC = CPUmem[SP];
	SP++;
	PC = PC | CPUmem[SP] << 8;
}

void RTI() {
	/* Not implemented */
}

void BRK() {
	/* NOP */
}

void JMP() {
	unsigned short address;
	address = CPUmem[PC++];
	address = address | CPUmem[PC++] << 8;
	PC = address;
}

void JSR() {
	unsigned short address;

	address = CPUmem[PC++];
	address = address | CPUmem[PC++] << 8;

	storeByte(SP,(PC >> 8)&0xFF);
	SP--;
	storeByte(SP,PC&0xFF);
	SP--;

	PC = address;
}

void UPPR0() {
	R0++;
	if (R0 == 0) {
		setCFlag(0x100);
		R1++;
	} else {
		setCFlag(0x00);
	}
}

void UPPR1() {
	R1++;
	if (R1 == 0) {
		setCFlag(0x100);
		R2++;
	} else {
		setCFlag(0x00);
	}
}

void UPPR2() {
	R2++;
	if (R2 == 0) {
		R3++;
		setCFlag(0x100);
	} else {
		setCFlag(0x00);
	}
}

void UPPR3() {
	R3++;
	if (R3 == 0) {
		R4++;
		setCFlag(0x100);
	} else {
		setCFlag(0x00);
	}
}

void UPPR4() {
	R4++;
	if (R4 == 0) {
		R5++;
		setCFlag(0x100);
	} else {
		setCFlag(0x00);
	}
}

void UPPR5() {
	R5++;
	if (R5 == 0) {
		R6++;
		setCFlag(0x100);
	} else {
		setCFlag(0x00);
	}
}

void UPPR6() {
	R6++;
	if (R6 == 0) {
		R7++;
		setCFlag(0x100);
	} else {
		setCFlag(0x00);
	}
}

void UPPR7() {
	R7++;
	if (R7 == 0) {
		R0++;
		setCFlag(0x100);
	} else {
		setCFlag(0x00);
	}
}

void STAR0() {
	unsigned short address;

	address = CPUmem[PC++];
	address = address | CPUmem[PC++] << 8;

	storeByte(address,R0);
}

void STAR1() {
	unsigned short address;

	address = CPUmem[PC++];
	address = address | CPUmem[PC++] << 8;

	storeByte(address,R1);
}

void STAR2() {
	unsigned short address;

	address = CPUmem[PC++];
	address = address | CPUmem[PC++] << 8;

	storeByte(address,R2);
}

void STAR3() {
	unsigned short address;

	address = CPUmem[PC++];
	address = address | CPUmem[PC++] << 8;

	storeByte(address,R3);
}

void STAR4() {
	unsigned short address;

	address = CPUmem[PC++];
	address = address | CPUmem[PC++] << 8;


	storeByte(address,R4);
}

void STAR5() {
	unsigned short address;

	address = CPUmem[PC++];
	address = address | CPUmem[PC++] << 8;

	storeByte(address,R5);
}

void STAR6() {
	unsigned short address;

	address = CPUmem[PC++];
	address = address | CPUmem[PC++] << 8;

	storeByte(address,R6);
}

void STAR7() {
	unsigned short address;

	address = CPUmem[PC++];
	address = address | CPUmem[PC++] << 8;

	storeByte(address,R7);
}

void STXR0() {
	unsigned short address;

	address = R0;
	address = address | R1 << 8;

	storeByte(address,R0);
}

void STXR1() {
	unsigned short address;

	address = R1;
	address = address | R2 << 8;

	storeByte(address,R0);
}

void STXR2() {
	unsigned short address;

	address = R2;
	address = address | R3 << 8;

	storeByte(address,R0);
}

void STXR3() {
	unsigned short address;

	address = R3;
	address = address | R4 << 8;

	storeByte(address,R0);
}

void STXR4() {
	unsigned short address;

	address = R4;
	address = address | R5 << 8;

	storeByte(address,R0);
}

void STXR5() {
	unsigned short address;

	address = R5;
	address = address | R6 << 8;

	storeByte(address,R0);
}

void STXR6() {
	unsigned short address;

	address = R6;
	address = address | R7 << 8;

	storeByte(address,R0);
}

void STXR7() {
	unsigned short address;

	address = R7;
	address = address | R0 << 8;

	storeByte(address,R0);
}

void STOR0() {
	unsigned short address;
	signed short offset;

	offset = CPUmem[PC++];

	address = R0;
	address = address | R1 << 8;

	address = address + offset;

	storeByte(address,R0);
}

void STOR1() {
	unsigned short address;
	signed short offset;

	offset = CPUmem[PC++];

	address = R1;
	address = address | R2 << 8;

	address = address + offset;

	storeByte(address,R0);
}

void STOR2() {
	unsigned short address;
	signed short offset;

	offset = CPUmem[PC++];

	address = R2;
	address = address | R3 << 8;

	address = address + offset;

	storeByte(address,R0);
}

void STOR3() {
	unsigned short address;
	signed short offset;

	offset = CPUmem[PC++];

	address = R3;
	address = address | R4 << 8;

	address = address + offset;

	storeByte(address,R0);
}

void STOR4() {
	unsigned short address;
	signed short offset;

	offset = CPUmem[PC++];

	address = R4;
	address = address | R5 << 8;

	address = address + offset;

	storeByte(address,R0);
}

void STOR5() {
	unsigned short address;
	signed short offset;

	offset = CPUmem[PC++];

	address = R5;
	address = address | R6 << 8;

	address = address + offset;

	storeByte(address,R0);
}

void STOR6() {
	unsigned short address;
	signed short offset;

	offset = CPUmem[PC++];

	address = R6;
	address = address | R7 << 8;

	address = address + offset;

	storeByte(address,R0);
}

void STOR7() {
	unsigned short address;
	signed short offset;

	offset = CPUmem[PC++];

	address = R7;
	address = address | R0 << 8;

	address = address + offset;

	storeByte(address,R0);
}

void LDIR0() {
	R0 = CPUmem[PC++];
	setNZFlag(R0);
}

void LDIR1() {
	R1 = CPUmem[PC++];
	setNZFlag(R1);
}

void LDIR2() {
	R2 = CPUmem[PC++];
	setNZFlag(R2);
}

void LDIR3() {
	R3 = CPUmem[PC++];
	setNZFlag(R3);
}

void LDIR4() {
	R4 = CPUmem[PC++];
	setNZFlag(R4);
}

void LDIR5() {
	R5 = CPUmem[PC++];
	setNZFlag(R5);
}

void LDIR6() {
	R6 = CPUmem[PC++];
	setNZFlag(R6);
}

void LDIR7() {
	R7 = CPUmem[PC++];
	setNZFlag(R7);
}

void LDAR0() {
	unsigned short address;
	address = CPUmem[PC++];
	address = address | CPUmem[PC++] << 8;
	R0 = loadByte(address);
	setNZFlag(R0);
}

void LDAR1() {
	unsigned short address;
	address = CPUmem[PC++];
	address = address | CPUmem[PC++] << 8;
	R1 = loadByte(address);
	setNZFlag(R1);
}

void LDAR2() {
	unsigned short address;
	address = CPUmem[PC++];
	address = address | CPUmem[PC++] << 8;
	R2 = loadByte(address);
	setNZFlag(R2);
}

void LDAR3() {
	unsigned short address;
	address = CPUmem[PC++];
	address = address | CPUmem[PC++] << 8;
	R3 = loadByte(address);
	setNZFlag(R3);
}

void LDAR4() {
	unsigned short address;
	address = CPUmem[PC++];
	address = address | CPUmem[PC++] << 8;
	R4 = loadByte(address);
	setNZFlag(R4);
}

void LDAR5() {
	unsigned short address;
	address = CPUmem[PC++];
	address = address | CPUmem[PC++] << 8;
	R5 = loadByte(address);
	setNZFlag(R5);
}

void LDAR6() {
	unsigned short address;
	address = CPUmem[PC++];
	address = address | CPUmem[PC++] << 8;
	R6 = loadByte(address);
	setNZFlag(R6);
}

void LDAR7() {
	unsigned short address;
	address = CPUmem[PC++];
	address = address | CPUmem[PC++] << 8;
	R7 = loadByte(address);
	setNZFlag(R7);
}

void LDXR0() {
	unsigned short address;

	address = R0;
	address = address | R1 << 8;
	R0 = loadByte(address);

	setNZFlag(R0);
}

void LDXR1() {
	unsigned short address;
	address = R1;
	address = address | R2 << 8;
	R0 = loadByte(address);
	setNZFlag(R0);
}

void LDXR2() {
	unsigned short address;
	address = R2;
	address = address | R3 << 8;
	R0 = loadByte(address);
	setNZFlag(R0);
}

void LDXR3() {
	unsigned short address;
	address = R3;
	address = address | R4 << 8;
	setNZFlag(R0);
}

void LDXR4() {
	unsigned short address;
	address = R4;
	address = address | R5 << 8;
	R0 = loadByte(address);
	setNZFlag(R0);
}

void LDXR5() {
	unsigned short address;
	address = R5;
	address = address | R6 << 8;
	R0 = loadByte(address);
	setNZFlag(R0);
}

void LDXR6() {
	unsigned short address;
	address = R6;
	address = address | R7 << 8;
	R0 = loadByte(address);
	setNZFlag(R0);
}

void LDXR7() {
	unsigned short address;
	address = R7;
	address = address | R0 << 8;
	R0 = loadByte(address);
	setNZFlag(R0);
}

void LDOR0() {
	unsigned short address;
	signed short offset;
	address = R0;
	address = address | R1 << 8;
	offset = CPUmem[PC++];
	address = address + offset;
	R0 = loadByte(address);
	setNZFlag(R0);
}

void LDOR1() {
	unsigned short address;
	signed short offset;
	address = R1;
	address = address | R2 << 8;
	offset = CPUmem[PC++];
	address = address + offset;
	R0 = loadByte(address);
	setNZFlag(R0);
}

void LDOR2() {
	unsigned short address;
	signed short offset;
	address = R2;
	address = address | R3 << 8;
	offset = CPUmem[PC++];
	address = address + offset;
	R0 = loadByte(address);
	setNZFlag(R0);
}

void LDOR3() {
	unsigned short address;
	signed short offset;
	address = R3;
	address = address | R4 << 8;
	offset = CPUmem[PC++];
	address = address + offset;
	R0 = loadByte(address);
	setNZFlag(R0);
}

void LDOR4() {
	unsigned short address;
	signed short offset;
	address = R4;
	address = address | R5 << 8;
	offset = CPUmem[PC++];
	address = address + offset;
	R0 = loadByte(address);
	setNZFlag(R0);
}

void LDOR5() {
	unsigned short address;
	signed short offset;
	address = R5;
	address = address | R6 << 8;
	offset = CPUmem[PC++];
	address = address + offset;
	R0 = loadByte(address);
	setNZFlag(R0);
}

void LDOR6() {
	unsigned short address;
	signed short offset;
	address = R6;
	address = address | R7 << 8;
	offset = CPUmem[PC++];
	address = address + offset;
	R0 = loadByte(address);
	setNZFlag(R0);
}

void LDOR7() {
	unsigned short address;
	signed short offset;
	address = R7;
	address = address | R0 << 8;
	offset = CPUmem[PC++];
	address = address + offset;
	R0 = loadByte(address);
	setNZFlag(R0);
}

void ILL() {
}

void (*opcodefuncs[256])() = {
	&INCR0,&INCR1,&INCR2,&INCR3,&INCR4,&INCR5,&INCR6,&INCR7,
	&ADCR0,&ADCR1,&ADCR2,&ADCR3,&ADCR4,&ADCR5,&ADCR6,&ADCR7,
	
	&TX0R0,&TX0R1,&TX0R2,&TX0R3,&TX0R4,&TX0R5,&TX0R6,&TX0R7,
	&ORR0,&ORR1,&ORR2,&ORR3,&ORR4,&ORR5,&ORR6,&ORR7,

	&ANDR0,&ANDR1,&ANDR2,&ANDR3,&ANDR4,&ANDR5,&ANDR6,&ANDR7,
	&XORR0,&XORR1,&XORR2,&XORR3,&XORR4,&XORR5,&XORR6,&XORR7,

	&ROLR0,&ROLR1,&ROLR2,&ROLR3,&ROLR4,&ROLR5,&ROLR6,&ROLR7,
	&RORR0,&RORR1,&RORR2,&RORR3,&RORR4,&RORR5,&RORR6,&RORR7,

	&DECR0,&DECR1,&DECR2,&DECR3,&DECR4,&DECR5,&DECR6,&DECR7,
	&SBCR0,&SBCR1,&SBCR2,&SBCR3,&SBCR4,&SBCR5,&SBCR6,&SBCR7,

	&ADDR0,&ADDR1,&ADDR2,&ADDR3,&ADDR4,&ADDR5,&ADDR6,&ADDR7,
	&STP0,&STP1,&STP2,&STP3,&STP4,&STP5,&STP6,&STP7,

	&BTT0,&BTT1,&BTT2,&BTT3,&BTT4,&BTT5,&BTT6,&BTT7,
	&CLP0,&CLP1,&CLP2,&CLP3,&CLP4,&CLP5,&CLP6,&CLP7,

	&T0XR0,&T0XR1,&T0XR2,&T0XR3,&T0XR4,&T0XR5,&T0XR6,&T0XR7,
	&CMPR0,&CMPR1,&CMPR2,&CMPR3,&CMPR4,&CMPR5,&CMPR6,&CMPR7,

	&PSHR0,&PSHR1,&PSHR2,&PSHR3,&PSHR4,&PSHR5,&PSHR6,&PSHR7,
	&POPR0,&POPR1,&POPR2,&POPR3,&POPR4,&POPR5,&POPR6,&POPR7,

	&BR00,&BR01,&BR02,&BR03,&BR04,&BR05,&BR06,&BR07,
	&BR10,&BR11,&BR12,&BR13,&BR14,&BR15,&BR16,&BR17,

	&USR,&ILL,&ILL,&ILL,&ILL,&ILL,&ILL,&ILL,
	&INT0,&INT1,&INT2,&INT3,&INT4,&INT5,&INT6,&INT7,

	&USR2,&ILL,&ILL,&ILL,&ILL,&ILL,&ILL,&ILL,
	&RSP,&RTS,&RTI,&BRK,&JMP,&ILL,&ILL,&JSR,
	
	&UPPR0,&UPPR1,&UPPR2,&UPPR3,&UPPR4,&UPPR5,&UPPR6,&UPPR7,
	&STAR0,&STAR1,&STAR2,&STAR3,&STAR4,&STAR5,&STAR6,&STAR7,

	&STXR0,&STXR1,&STXR2,&STXR3,&STXR4,&STXR5,&STXR6,&STXR7,
	&STOR0,&STOR1,&STOR2,&STOR3,&STOR4,&STOR5,&STOR6,&STOR7,

	&LDIR0,&LDIR1,&LDIR2,&LDIR3,&LDIR4,&LDIR5,&LDIR6,&LDIR7,
	&LDAR0,&LDAR1,&LDAR2,&LDAR3,&LDAR4,&LDAR5,&LDAR6,&LDAR7,

	&LDXR0,&LDXR1,&LDXR2,&LDXR3,&LDXR4,&LDXR5,&LDXR6,&LDXR7,
	&LDOR0,&LDOR1,&LDOR2,&LDOR3,&LDOR4,&LDOR5,&LDOR6,&LDOR7
};

void initCPUstate(char *filename) {
	FILE *f;
	int totalfound;
	int totalsize;
	int found;

	totalfound = 0;

	f = fopen(filename,"rb");
	if (!f) {
		printf("Error opening file %s\n",filename);
		exit(1);
	}

	// Load in the fixed bank.
	found = fread(&CPUmem[0x090],1,0x1000,f);
	totalfound += found;

	//  Load in the other banks.
	found = fread(bank0,1,3*0x1000,f);
	totalfound += found;

	found = fread(bank1,1,3*0x1000,f);
	totalfound += found;

	found = fread(bank2,1,3*0x1000,f);
	totalfound += found;

	found = fread(bank3,1,3*0x1000,f);
	totalfound += found;

	found = fread(bank4,1,3*0x1000,f);
	totalfound += found;

	found = fread(bank5,1,3*0x1000,f);
	totalfound += found;

	found = fread(bank6,1,3*0x1000,f);
	totalfound += found;

	found = fread(bank7,1,3*0x1000,f);
	totalfound += found;

	found = fread(bank8,1,3*0x1000,f);
	totalfound += found;

	found = fread(bank9,1,3*0x1000,f);
	totalfound += found;

	fseek(f,0L,SEEK_END);
	totalsize = ftell(f);
	fclose(f);

	// Copy in bank 0
	memcpy(&CPUmem[0x1000],bank0,3*0x1000);

	init_CPU();
	PC = 0x0090;
}

void printmemline(unsigned int location,char *outline) {
	int count;

	sprintf(outline,"%04X: ",location&0xFFFF);
	for (count=0;count<16;count++) {
		sprintf(outline+6+(count*3),"%02X ",CPUmem[(location+count)&0xFFFF]);
	}
	
	sprintf(outline+6+16*3,"    ");

	for(count=0;count<16;count++) {
		if(CPUmem[(location+count)&0xFFFF] < 0x20) {
			sprintf(outline+6+16*3+3+count,"?");
		} else {
			sprintf(outline+6+16*3+3+count,"%c",CPUmem[(location+count)&0xFFFF]);
		}
	}
}

void doCPUstep(char *CPUoutline,char *opcodeoutline) {
	unsigned char opcode;
	int count;

	sprintf(CPUoutline,"%04X  %04X   %02X    %02X    %02X    %02X    %02X    %02X    %02X    %02X   ",PC,SP,R0,R1,R2,R3,R4,R5,R6,R7);

	for (count=0;count<8;count++) {
		if ( (PSR & (1 << count)) ) {
			sprintf(CPUoutline+60+count,"1");
		} else {
			sprintf(CPUoutline+60+count,"0");
		}
	}

	if (noopcode[CPUmem[PC]] == 1) {
		sprintf(opcodeoutline,"0x%04X 0x%02X\t\t%s",PC,CPUmem[PC],opcodes[CPUmem[PC]]);
	} else if (noopcode[CPUmem[PC]] == 2) {
		sprintf(opcodeoutline,"0x%04X 0x%02X 0x%02X\t%s #$%02X",PC,CPUmem[PC],CPUmem[PC+1],opcodes[CPUmem[PC]],CPUmem[PC+1]);
	} else {
		sprintf(opcodeoutline,"0x%04X 0x%02X 0x%02X 0x%02X\t%s $%02X%02X",PC,CPUmem[PC],CPUmem[PC+1],CPUmem[PC+2],opcodes[CPUmem[PC]],CPUmem[PC+2],CPUmem[PC+1]);
	}

	// Execute the instruction.
	opcode = CPUmem[PC++];
	( *(opcodefuncs[opcode]) )();
}
