#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <curses.h>

#include "V8sim.h"
/*
 * LINES and COLS contain the rows and cols.
 */

void exitcurses() {
	clear();
	printw("Press any key to exit.");
	refresh();
	cbreak();
	noecho();
	getch();	
	endwin();
}

void my_clear() {
	int count;
	for (count=0;count<LINES*COLS;count++) {
		printw(" ");
	}
	box(stdscr,0,0);
	move(0,2);
	printw("mem watch");
	move(11,1);
	hline(0,COLS-2);
	move(11,2);
	printw("  PC    SP    R0    R1    R2    R3    R4    R5    R6    R7   ZCNI----    ");
	move(13,1);
	hline(0,COLS-2);
	move(13,2);
	printw("disasm output");
	move(LINES-3,1);
	hline(0,COLS-2);
	move(LINES-3,2);
	printw("command");
}

void printmem(int location) {
	int count;
	char myline[80];

	for(count=1;count<11;count++) {
		memset(myline,0,80);
		printmemline(location+((count-1)*16),myline);
		move(count,1);
		printw(myline);
	}
	refresh();
}

void initcurses() {
	initscr();
	noecho();
	cbreak(); /* Line buffering disabled */
	start_color();
	init_pair(1,COLOR_YELLOW,COLOR_BLUE);
	attrset(COLOR_PAIR(1) | A_BOLD);
	my_clear();
	init_pair(2,COLOR_WHITE,COLOR_BLUE);
	attrset(COLOR_PAIR(2));
	refresh();
}

char disasmlines[8][80];

void drawdisasm(char *cpuline) {
	int count;

	for(count=0;count<7;count++) {
		memcpy(&disasmlines[count][0],&disasmlines[count+1][0],80);
	}
	memcpy(&disasmlines[7][0],cpuline,80);

	for(count=0;count<8;count++) {
		move(14+count,1);
		printw("                                       ");
		move(14+count,1);
		printw(&disasmlines[count][0]);
	}
} 

int mempointer;

void dostep(int howmany) {
	int count;
	char CPUoutline[80];
	char opcodeoutline[80];

	for (count=0;count<howmany;count++) {
		memset(CPUoutline,0,80);
		memset(opcodeoutline,0,80);

		printmem(mempointer&0xFFFF);
		doCPUstep(CPUoutline,opcodeoutline);
		move(12,3);
		printw(CPUoutline);
		drawdisasm(opcodeoutline);	
		move(LINES-2,1);
		printw("                                       ");
		move(LINES-2,1);
		refresh();
	}
}

int main(int argc,char **argv) {
	char cmd[80];
	int cmdpointer;
	int done;
	char read;
	int nosteps;

	cmdpointer = 0;

	if (argc < 2) {
		printf("Usage: V8sim <filename>\n");
		exit(1);
	}

	initCPUstate(argv[1]);

	/*
         * First interesting location of the PV2 code.
	 * Also interesting: F800 / F900 SRAM <-> SDRAM transfer.
         */
	mempointer = 0x0DF0;

	initcurses();
	dostep(1);

	done = 0;


	while(!done) {
		switch(read = getch()) {
			case 27:	done = 1;
					break;

			case ' ':
			case 'm':
			case 's':
			case 'r':
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
			case 'a':
			case 'b':
			case 'c':
			case 'd':
			case 'e':
			case 'f':
			case 'A':
			case 'B':
			case 'C':
			case 'D':
			case 'E':
			case 'F':	printw("%c",read);
					cmd[cmdpointer++] = read;
					refresh();
					break;
			
			case 8:		if (cmdpointer > 0) {
						move(LINES-2,cmdpointer);
						printw(" ");
						move(LINES-2,cmdpointer);
						cmdpointer--;
						cmd[cmdpointer] = 0;
						refresh();
					}
					break;
					

			case 10:	if ((cmdpointer == 0) || (cmd[0] == 's')) {
						dostep(1);
					} else if (cmd[0] == 'r') {
						sscanf(cmd,"r %d",&nosteps);
						dostep(nosteps);
					} else if (cmd[0] == 'm') {
						sscanf(cmd,"m %04X",&mempointer);
						printmem(mempointer);
					}
					move(LINES-2,1);
					printw("                                       ");
					move(LINES-2,1);
					refresh();
					cmdpointer = 0;
					memset(cmd,0,80);
					break;

			default:	/* Not supported */
					break;
		}
	}
	exitcurses();

	return(0);
}
